#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Dendrogramix
a Hybrid Tree-Matrix Visualization Technique
to Support Interactive Exploration of Dendrograms

implementation of the prototype described in [Blanch et al., 2015]

[Blanch et al., 2015]
Dendrogramix: a Hybrid Tree-Matrix Visualization Technique to Support Interactive Exploration of Dendrograms
Renaud Blanch, Rémy Dautriche, Gilles Bisson 
In Proceedings of the 8th IEEE Pacific Visualization Symposium (PacificVis 2015). pages 31-38. 2015.
http://iihm.imag.fr/blanch/projects/dendrogramix/

Copyright (c) 2011--2015, IIHM/LIG - Renaud Blanch <http://iihm.imag.fr/blanch/>
Licence: GPLv3 or higher <http://www.gnu.org/licenses/gpl.html>
"""


#############################################################################
# data set ##################################################################
#############################################################################

import sys
import os


from collections import defaultdict

# loading ###################################################################
#############################################################################

DATA_FOLDER = os.path.join(os.path.dirname(sys.argv[0]), "../data")

#dataset = 'IV04_DATA'
dataset = 'InfoVis_2006_2011_DATA'
#dataset = 'LIG_2008_2012_DATA'


# IV04 data set #############################################################

if dataset == 'IV04_DATA': 
	iv04_folder = os.path.join(DATA_FOLDER, "iv04dataset-final-05-05-2004/Additional materials provided by other paricipants")

	dup_authors = open(os.path.join(iv04_folder, "duplicate_author_map.txt"))
	unique_authors = dict(tuple(w.split(",")[0][1:]
	                            for w in line.strip().split(" -> "))
	                      for line in dup_authors if line.strip())

	publication_authors = defaultdict(list)
	publications = open(os.path.join(iv04_folder, "tableform-of dataset(dup-authors-not-applied)/articles_by_author.txt"))
	author_names = dict()

	next(publications) # skip header line
	for line in publications:
		article, _, author_id, name = line.strip().split(None, 3)
		try:
			author_id = unique_authors[author_id]
		except KeyError:
			author_names[author_id] = name.decode('iso-8859-1')
		publication_authors[article].append(author_id)


# InfoVis 2006-2011 data set ################################################

if dataset == 'InfoVis_2006_2011_DATA':
	iv11_folder = os.path.join(DATA_FOLDER, "infovis", "2006-2011")

	# author_names
	author_names = {}
	unique_ids = {}
	for line in open(os.path.join(iv11_folder, "unique_authors.txt")):
		author_id, unique_id, author_name = line.strip().split(None, 2)
		author_id, unique_id = int(author_id), int(unique_id)
		unique_ids[author_id] = unique_id
		author_names[unique_id] = author_name.decode("utf-8")
	
	# publications
	publication_authors = defaultdict(list)
	for line in open(os.path.join(iv11_folder, "publications.txt")):
		pub_id, author_id = (int(u) for u in line.strip().split())
		publication_authors[pub_id].append(author_names[unique_ids[author_id]])


# LIG co-authors 2008-2012 data set #########################################

if dataset == 'LIG_2008_2012_DATA':
	import json
	
	lig_folder = os.path.join(DATA_FOLDER, "pistou", "2013-12-11")

	lig_authors = json.load(open(os.path.join(lig_folder, "perms.json")))
	publications = json.load(open(os.path.join(lig_folder, "lig.json")))["publications"]

	def author_key(author_dict):
		return u" ".join(author_dict[k] for k in ["nom", "prenom"])
	
	publication_authors = defaultdict(list)
	for i, pub in enumerate(publications):
		for author in pub["auteurs"]:
			author = author_key(author)
			if author not in lig_authors:
				continue
			author = "%s [%s]" % (author, ", ".join(lig_authors[author]))
			publication_authors[i].append(author)


# common prepocessing #######################################################

coauthor_count = defaultdict(int)
author_count = defaultdict(int)
for article in publication_authors:
	authors = publication_authors[article]
	for author in authors:
		author_count[author] += 1
		for coauthor in authors:
			coauthor_count[author, coauthor] += 1

author_order = list(author
                    for author in author_count if author_count[author] > 2)
author_score = list(list(coauthor_count[author, coauthor]
                         for coauthor in author_order)
                    for author in author_order)

names = author_order
vectors = author_score


# clustering ################################################################
#############################################################################

from scipy.spatial.distance import pdist, squareform
from scipy.cluster.hierarchy import linkage, leaves_list


metric = 'cosine'
#metric = 'correlation'
method = 'single'
#method = 'average'
#method = 'weighted'
#method = 'complete'

y = pdist(vectors, metric=metric)
Z = linkage(y, method=method, metric=metric)
Z = [(int(l), int(r), max(0., s), int(n)) for (l, r, s, n) in Z] # cleaning

leaves = list(leaves_list(Z))
count = len(leaves)
root = len(Z)+count-1

X = squareform(y)
assert len(X) == count


from utils import memoise

@memoise
def single(c, d):
	return min(min(X[lc][ld] for lc in leaves if lc in childrens[c])
	           for ld in leaves if ld in childrens[d])


# bar-joseph optimal ordering ###############################################

OPTIMAL_ORDER = True
if OPTIMAL_ORDER:
	from barjoseph import optimal
	from math import exp
	
	leaves = optimal(root, **{
		"S":        lambda i, j: exp(-X[i][j]),
		"left":     lambda i: None if i < count else Z[i-count][0],
		"right":    lambda i: None if i < count else Z[i-count][1],
		"is_leaf":  lambda i: i < count,
		"is_empty": lambda v: v is None,
	})


assert list(sorted(leaves)) == range(count)


# cluster helpers ###########################################################

def update_spans():
	"""given a leaves order, compute spans
	
	spans[c]   is (i_min, i_max), the bounds of cluster c
	"""
	global spans
	spans = list((leaves.index(i),) * 2 for i in range(count))
	for l, r, _, _ in Z:
		l_min, l_max = spans[l]
		r_min, r_max = spans[r]
		spans.append((min(l_min, r_min), max(l_max, r_max)))

def init_clusters():
	"""precompute parents, children and spans.

	parents[c]   is the parent cluster of c
	childrens[c] is the list of c children
	labels[c]    is the label of c
	"""
	global parents, childrens, labels
	parents = range(root+1)
	childrens = [[i] for i in range(count)]
	for (l, r, _, _), c in zip(Z, range(count, count+len(Z))):
		parents[l] = parents[r] = c
		childrens.append([c] + childrens[l] + childrens[r])
	labels = {}
	update_spans()

init_clusters()


def width(c):
	c_min, c_max = spans[c]
	return c_max - c_min + 1

def lr(c):
	"""left/right child of c."""
	k = c - count
	if k < 0: # leaf
		return c, c
	l, r = Z[k][:2]
	if spans[l] < spans[r]:
		return l, r
	else:
		return r, l


def flip(c):
	"""flips c node"""
	global leaves, spans
	l, r = lr(c)
	if l == r: # leaf
		return
	l_min, l_max = spans[l]
	r_min, r_max = spans[r]
	assert l_max+1 == r_min
	leaves[l_min:r_max+1] = leaves[r_min:r_max+1]+leaves[l_min:l_max+1]
	folding_factors[l_min:r_max+1] = folding_factors[r_min:r_max+1]+folding_factors[l_min:l_max+1]
	update_spans()
	update_fold_map()
	
	
def right_child_parent(c):
	"""smallest englobing cluster which is a right child of its parent."""
	_, c_max = spans[c]
	while c < root:
		c = parents[c]
		_, p_max = spans[c]
		if p_max == c_max:
			break
		c_max = p_max
	return c


def left_child_parent(c):
	"""smallest englobing cluster which is a left child of its parent."""
	c_min, _ = spans[c]
	while c < root:
		c = parents[c]
		p_min, _ = spans[c]
		if p_min == c_min:
			break
		c_min = p_min
	return c


#############################################################################
# graphics ##################################################################
#############################################################################

from OpenGL.GLUT import *
from OpenGL.GL import *

from time import time
from math import *

SQRT22 = sqrt(2.)/2. # = sin(pi/4.)


WINDOW_SIZE = None
WIDTH, HEIGHT = 1600, 1200

SIZE = 2.5
HIGHLIGHT_RGB = .0, 0., .75
NAME_LEN = max(len(name) for name in names)

similarity_threshold = 1. # threshold on values in similarity matrix
cluster_threshold = .5    # threshold on clustering cut

current = root
highlight = [0, count-1]

bending = 200.
foldeds = []
folding_factors = [1] * count
folding_widths = {}

def update_fold_map():
	global fold_map
	fold_map = list(sum(1./folding_factors[i] for i in range(n))
	                for n in range(count+1))
update_fold_map()


# display ###################################################################
#############################################################################

from copy import deepcopy

def displaylist(*state):
	"""wraps a drawing function into a display list.
	
	tracks state changes to invalidate the display list
	"""
	
	def decorator(draw):
		draw._list_id = None
		def reset():
			if draw._list_id != None:
				glDeleteLists(draw._list_id, 1)
			draw._list_id = None
		
		draw._state = []
		def _draw(*args, **kwargs):
			if state != draw._state:
				draw._state = deepcopy(state)
				reset()
				
			if draw._list_id is None:
				list_id = glGenLists(1)
				glNewList(list_id, GL_COMPILE_AND_EXECUTE)
				draw(*args, **kwargs)
				glEndList()
				draw._list_id = list_id
			else:
				glCallList(draw._list_id)
		
		_draw.__name__ = draw.__name__
		_draw.__doc__  = draw.__doc__
		_draw._reset = reset
		return _draw
	
	return decorator


# low level #################################################################

def line_points(u):
	for k in range(u):
		yield (u, k)
	for k in range(u, 2*count+1):
	 	yield (k, u)

def line(u):
	"""line (row/col) for coordinate u.
	
	u == 2*i     -> left border of observation i
	u == 2*i + 1 -> middle of i
	u == 2*(i+1) -> right border of i
	"""
	glBegin(GL_LINE_STRIP)
	for p in line_points(u):
		glVertex(*p)
	glEnd()


def cluster_points(c):
	"""outline of cluster c."""
	c_min, c_max = spans[c]
	for k in range(2*c_min, 2*(c_max+1)+1):
		yield (2*(c_max+1), k)
	for k in reversed(range(2*c_min, 2*(c_max+1))):
		yield (k, k)
	for k in range(2*c_min, 2*(c_max+1)):
		yield (k, 2*c_min)

def cluster(c, fill, fill_alpha=1., stroke=0., stroke_alpha=.2):
	for color, MODE in [((fill,   fill,   fill,   fill_alpha),   GL_TRIANGLE_FAN),
	                    ((stroke, stroke, stroke, stroke_alpha), GL_LINE_LOOP)]:
		glColor(*color)
		glBegin(MODE)
		for p in cluster_points(c):
			glVertex(*p)
		glEnd()


def dot(c, d, s):
	c_min, c_max = spans[c]
	d_min, d_max = spans[d]
	r = sqrt(acos(s))/1.5
#	for c, r in [((0., 0., 0.), r+.2), ((s, s, s), r)]:
#	for c, r in [((0., 0., 0.), r+.2), ((.5, .5, .5), r)]:
	for c, r in [((.5, .5, .5), r+.2), ((0., 0., 0.), r)]:
		ri = 1./r
		glColor(*c)
		glTexCoord(-ri, -ri)
		glVertex(2*c_min, 2*d_min)
		glTexCoord(-ri, ri)
		glVertex(2*c_min, 2*(d_max+1))
		glTexCoord(ri, ri)
		glVertex(2*(c_max+1), 2*(d_max+1))
		glTexCoord(ri, -ri)
		glVertex(2*(c_max+1), 2*d_min)


def author(name, i, s=1., dl=0):
	glPushMatrix()
	glTranslate(2*i, 2*i, 0.)
	glColor(s, s, s)
	text_v(NBSP, -3-SIZE*3)
	glColor(0., 0., 0.)
	text_v(name, -6*(len(name)-dl+1)-SIZE*3)
	glPopMatrix()
	

# data ######################################################################

@displaylist(leaves)
def draw_clusters():
	# clusters
	for i, (_, _, s, _) in reversed(list(enumerate(Z))):
		c = count + i
		cluster(c, s)
	
	# grid
#	glColor(0., 0., 0., .1)
#	for i in range(count+1):
#		line(2*i)

@displaylist(leaves)
def draw_values():
	clusters = [c for c in range(count) + foldeds
	            if not any(c in childrens[f] for f in foldeds if c != f)]
	glUniform1i(locations[b"circle"], True)
	glBegin(GL_QUADS)
	for c in clusters:
		c_min, c_max = spans[c]
		for d in clusters:
			d_min, d_max = spans[d]
			if d_max > c_min:
				continue
			s = single(c, d)
			if s >= similarity_threshold:
				continue
			dot(c, d, s)
	glEnd()
	glUniform1i(locations[b"circle"], False)


@displaylist(leaves, foldeds, labels)
def draw_labels():
	glDisable(GL_MULTISAMPLE)
	glColor(0., 0., 0.)
	for c in labels:
		for folded in reversed(foldeds):
			if c in childrens[folded]:
				c = None
				break
		if c is None:
			continue
		c_min, c_max = spans[c]
		label = labels[c]
		if Z[c-count][2] < .75:
			bg, cg, fg = 0., .75, 1.
		else:
			bg, cg, fg = 1., .25, 0.
		l = max(1, len(label))
		l3 = l*3
		glPushMatrix()
		glTranslate(2*c_max+1, 2*c_min+1, 0.)
		glColor(bg, bg, bg, .75)
		if interaction == LABELING and c == current:
			text_h(NBSP*l, -l3+1, -10)
			glColor(cg, cg, cg)
			text_h(SHY, -l3+1+6*(caret-1), -10)
		else:
			text_h(label, -l3, -11)
		glColor(fg, fg, fg)
		text_h(label, -l3+1, -10)
		glPopMatrix()
	glEnable(GL_MULTISAMPLE)


@displaylist(leaves, foldeds, labels, highlight)
def draw_authors():
	glDisable(GL_MULTISAMPLE)
	for i, l in enumerate(leaves):
		if any(l in childrens[f] for f in foldeds):
			continue
		name = names[l]
		x = X[l]
		s = 1. - sqrt((1.-x[leaves[highlight[0]]])*(1.-x[leaves[highlight[1]]]))
		author(name, i, s)
	
	for folded in foldeds:
		if any(folded in childrens[f] for f in foldeds if f != folded):
			continue
		name = labels.get(folded, u"") or str(folded)
		name = u"[%s]»" % name
		for h in highlight:
			l = leaves[h]
			if l in childrens[folded]:
				name = u"".join([names[l][:-len(name)-2], unichr(133), " ", name])
				break
		i, _ = spans[folded]
		author(name, i, dl=1)
	glEnable(GL_MULTISAMPLE)


# feedback ##################################################################
		
def draw_drag():
	if di == 0.:
		return

	l, r = lr(parent)
	wl, wr = width(l), width(r)
	if di > 0.:
		if di < wr/2.:
			il, jl = di*2, 0.
			ir, jr = 0., 0.
		else:
			il, jl = wr, di*2.-wr
			ir, jr = -wl, -wl
	else:
		if -di < wl/2.:
			il, jl = 0., 0.
			ir, jr = 0., di*2.
		else:
			il, jl = wr, wr
			ir, jr = di*2.+wl, -wl
	
	cluster(parent, Z[parent-count][2])
	for c, i, j in [(l, il, jl), (r, ir, jr)]:
		glPushMatrix()
		glTranslate(2*i, 2*j, 0.)
		children = childrens[c]
		for k in children:
			if k > count:
				cluster(k, Z[k-count][2])
		if current in children:
			cluster(current, 1., 0., 0., .75)
		glPopMatrix()


def highlight_author(h, r, g, b):
	glColor(r, g, b, .25)
	glPushMatrix()
	glTranslatef(2*h, 2*h, 0.)
	text_v((NAME_LEN+1)*NBSP, -6*(NAME_LEN+3))
	glPopMatrix()

def highlight_authors(c, r, g, b):
	for f in reversed(foldeds):
		if c not in childrens[f]:
			continue
		f_min, _ = spans[f]
		return highlight_author(f_min, r, g, b)

	if c in leaves:
		return highlight_author(leaves.index(c), r, g, b)
	
	for u in lr(c):
		highlight_authors(u, r, g, b)


def draw_selection():
	for ((r, g, b), u) in zip([(.75, 0., 0.), (0., .5, 0.)],
	                          selection_clusters):
		u_min, u_max = spans[u]
		i, a = 2*u_min, 2*(u_max+1)
		glColor(r, g, b, .75)
		line(i)
		line(a)
		
		glColor(r, g, b, .25)
		glBegin(GL_TRIANGLE_FAN)
		for k in range(i, a+1):
			glVertex(k, 0)
		glEnd()
		glBegin(GL_TRIANGLE_FAN)
		for k in range(i, a+1):
			glVertex(2*count+1, k)
		glEnd()
		glBegin(GL_TRIANGLE_STRIP)
		for pi, pa in zip(line_points(i), line_points(a)):
			ii, ji = pi
			ia, ja = pa
			if ji != ja:
				if ii != ia:
					continue
				else:
					pa, pi = pi, pa
			glVertex(pi)
			glVertex(pa)
		glEnd()
		
		highlight_authors(u, r, g, b)


def draw_trace():
	r, g, b = HIGHLIGHT_RGB
	
	# selection trace
	glColor(r, g, b, .75)
	glUseProgram(0)
	glMatrixMode(GL_PROJECTION)
	glPushMatrix()
	glLoadIdentity()
	glOrtho(0, WIDTH, HEIGHT, 0, 0., 1.)

	glBegin(GL_LINE_STRIP)
	for p in selection_trace:
		glVertex(p)
	glEnd()
	glPopMatrix()
	glMatrixMode(GL_MODELVIEW)

	glUseProgram(program)


def draw_highlight():
	r, g, b = HIGHLIGHT_RGB

	cluster(current, 1., 0., 0., .75)
	highlight_authors(current, .5, .5, .5)
	
	for h in highlight:
		glColor(r, g, b, .75)
		line(2*h+1)
		highlight_authors(leaves[h], r, g, b)


# cluster thresholding ######################################################

def cluster_s(x, y):
	x = x-WIDTH/2
	y = 4*HEIGHT/5.-y
	x /= SIZE; y /= SIZE
	x -= 11-2.*fold_map[count]; y -= 120
	x /= 2.; y /= 2.
	x /= 1.; y /= 20.
	x -= 1
	x -= 1
	x *= count/40.
	return x, y

def draw_slider():
	glUseProgram(0)
	
	glPushMatrix()
	
	glTranslate(12-2.*fold_map[count], 120, 0.)
	glScale(2., 2., 1.)
	glScale(1., 20., 1.)
	glTranslate(1, 0, 0)
	
	for c, GL_TYPE in [(1., GL_QUADS), (0., GL_LINE_LOOP)]:
		glBegin(GL_TYPE)
		glColor(0., 0., 0.)
		glVertex(0., 0.)
		glVertex(-1., 0.)
		glColor(c, c, c)
		glVertex(-1., 1.)
		glVertex(0., 1.)
		glEnd()
	
	def x(k): return 40.*k/count+1
#	def x(k): return 16.*log(k+1, 10)/log(count+1, 10)
	
	glBegin(GL_LINE_STRIP)
	for n, (_, _, s, _) in enumerate(reversed(Z)):
		glVertex(x(n+1), s)
		if s >= cluster_threshold:
			k = n+1
	glEnd()
	
	glBegin(GL_LINE_STRIP)
	glVertex(-1.5, cluster_threshold)
	glVertex(x(k), cluster_threshold)
	glVertex(x(k), 0)
	glEnd()
	
	glDisable(GL_MULTISAMPLE)
	glPushMatrix()
	glTranslate(0., cluster_threshold, 0.)
	text_h("%.02f" % cluster_threshold, -6*6, -5)
	glPopMatrix()
	
	glPushMatrix()
	glTranslate(x(k), 0., 0.)
	k = "%s" % k
	text_h("1 cluster" if k == "1" else k + " clusters", -6*len(k)-3, -15)
	glPopMatrix()
	glEnable(GL_MULTISAMPLE)
	
	glPopMatrix()
	
	glUseProgram(program)


def threshold(c=root):
	if c < count:
		return []
	l, r, s, _ = Z[c-count]
	if s <= cluster_threshold:
		return [c]
	return threshold(l) + threshold(r)

@displaylist(leaves)
def draw_threshold():
	clusters = threshold()
	for line_width, stroke, stroke_opacity in [(4, 1., .75), (2, 0., 1.)]:
		glLineWidth(line_width)
		for c in clusters:
			cluster(c, 0., 0., stroke, stroke_opacity)
	glLineWidth(1)
	

# glut callbacks for display ################################################

def display():
	"""window redisplay callback."""
	glClear(GL_COLOR_BUFFER_BIT)
	
	draw_clusters()

	if interaction in [IDLE, CLICKING, FOLDING, SELECTING, LABELING, THRESHOLDING]:
		draw_threshold()
		draw_values()
		draw_authors()

		draw_selection()
	
		if interaction is SELECTING:
			draw_trace()
		
		if interaction in [IDLE, CLICKING, LABELING]:
			draw_highlight()

		draw_labels()
	
	elif interaction is DRAGGING:
		draw_drag()
		draw_selection()
	
	elif interaction is DRAGGING_TWO:
		draw_selection()
	
	draw_slider()
	
	glutSwapBuffers()


def reshape(width, height):
	"""window reshape callback."""
	global WIDTH, HEIGHT
	WIDTH, HEIGHT = width, height

	glViewport(0, 0, WIDTH, HEIGHT)

	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	glOrtho(0, WIDTH, 0, HEIGHT, 0, 1.)
	glTranslate(WIDTH/2, HEIGHT/5, 0.)
	glScale(SIZE, SIZE, 1.)
	
	glMatrixMode(GL_MODELVIEW)
	glLoadIdentity()


#############################################################################
# interaction ###############################################################
#############################################################################

def unfold(u):
	if u < 0:
		return u
	
	u_max = fold_map[count]
	if u > u_max:
		return u-u_max+count+1
	
	i = int(floor(u))
	fi0 = fi = fold_map[i]
	while fi <= u:
		i += 1
		fi0, fi = fi, fold_map[i]
	
	return i-1+(u-fi0)/(fi-fi0)
	
	
def ij(x, y):
	"""picking."""
	x = (x-WIDTH/2)/SIZE+2*fold_map[count]
	y = max(0., 4*HEIGHT/5.-y)/SIZE
	y = bending * (exp(y/bending) - 1.)
	return [unfold(u/4.) for u in [x-y, x+y]]

def leaves_index(x, y):
	return [min(max(0, int(h)), count-1) for h in ij(x, y)]

def cd(x, y):
	return [leaves[u] for u in leaves_index(x, y)]


IDLE, CLICKING, DRAGGING, DRAGGING_TWO, \
FOLDING, SELECTING, LABELING, THRESHOLDING = range(8)

interaction = IDLE


# mouse interaction #########################################################
#############################################################################

def mouse(button, state, x, y):
	global interaction, x0, y0
	
	if button == GLUT_LEFT_BUTTON:
		if state == GLUT_DOWN:
			if interaction is IDLE:
				if thresholding(x, y):
					interaction = THRESHOLDING
				else:
					interaction = CLICKING
		
		elif state == GLUT_UP:
			if interaction is CLICKING:
				global T_FOLD
				if glutGetModifiers() & GLUT_ACTIVE_SHIFT:
					T_FOLD = 5000
				else:
					T_FOLD = 500
				if current in foldeds:
					start_unfold()
				else:
					start_fold()
				interaction = FOLDING
			
			elif interaction is DRAGGING:
				end_drag()
				interaction = IDLE

			elif interaction is DRAGGING_TWO:
				end_drag_two()
				interaction = IDLE

			elif interaction is THRESHOLDING:
				interaction = IDLE
	
	elif button == GLUT_RIGHT_BUTTON:
		if state == GLUT_DOWN:
			if interaction is IDLE:
				start_select(x, y)
				interaction = SELECTING
		elif state == GLUT_UP:
			assert interaction is SELECTING
			end_select(x, y)
			interaction = IDLE
			
	x0, y0 = x, y
	glutPostRedisplay()


def motion(x, y):
	global interaction
	
	if interaction is IDLE:
		hover(x, y)
	
	elif interaction is THRESHOLDING:
		thresholding(x, y)
	
	elif interaction is CLICKING:
		if selection_clusters and \
		   all(c in childrens[d]
		       for (c, d) in zip(cd(x0, y0), selection_clusters)):
			interaction = DRAGGING_TWO
		else:
			interaction = DRAGGING
		motion(x, y)
	
	elif interaction is DRAGGING:
		drag(x, y)
	
	elif interaction is DRAGGING_TWO:
		drag_two(x, y)

	elif interaction is SELECTING:
		select(x, y)

	glutPostRedisplay()


# thresholding ##############################################################

def thresholding(x, y):
	global cluster_threshold
	
	c, s = cluster_s(x, y)
	if not (c <= count+10 and -.5 <= s <= 1.1):
		return False
	c = min(max(1, int(c)), count-1)
	s = min(max(0., s), 1.)

	if s > 1.*c/count:
		cluster_threshold = s
	else:
		cluster_threshold = Z[-c][2]
	draw_threshold._reset()
	
	return True


# selection #################################################################

def start_select(x0, y0):
	global selection_trace, selection_clusters
	if not glutGetModifiers() & GLUT_ACTIVE_SHIFT:
		deselect()
	
	selection_trace = [(x0, y0)]
	if selection_clusters == []:
		c, d = cd(x0, y0)
		selection_clusters = c, d
	else:
		select(x0, y0)


def select(x, y):
	global selection_trace, selection_clusters
	selection_trace.append((x, y))
	L, R = lr(current)
	l, r = selection_clusters
	c, d = cd(x, y)
	while c not in childrens[l]:
		if l == L or l not in childrens[L]:
			break
		l = parents[l]
	while d not in childrens[r]:
		if r == R or r not in childrens[R]:
			break
		r = parents[r]
	selection_clusters = l, r

def end_select(x, y):
	hover(x, y)
	
def deselect():
	global selection_clusters
	selection_clusters = []

deselect()


# folding ###################################################################

_START = time()
def ms():
	return int((time()-_START)*1000)

T_FOLD = 500 #ms

def start_fold():
	global f_min, f_max
	f_min, f_max = spans[current]
	glUniform1f(locations[b"f_min"], 2*f_min)
	glUniform1f(locations[b"f_max"], 2*(f_max+1))
	folding_widths[current] = fold_map[f_max+1]-fold_map[f_min]
	glutTimerFunc(0, _fold, ms()+T_FOLD)

def _fold(t_end):
	folding = 1.*(t_end - ms())/T_FOLD
	w = folding_widths[current]
	if folding*w < 1:
		global interaction
		interaction = IDLE
		folding = 1.
		foldeds.append(current)
		for c in range(f_min, f_max+1):
			folding_factors[c] *= w
		update_fold_map()
		glUniform1fv(locations[b"fold_map"], count+1, fold_map)
		draw_values._reset()
	else:
		glutTimerFunc(20, _fold, t_end)
	glUniform1f(locations[b"folding"], folding)
	glutPostRedisplay()


def start_unfold():
	global f_min, f_max
	f_min, f_max = spans[current]
	foldeds.remove(current)
	w = folding_widths[current]
	for c in range(f_min, f_max+1):
		folding_factors[c] /= w
	update_fold_map()
	glUniform1f(locations[b"folding"], 1./w)
	glUniform1f(locations[b"f_min"], 2*f_min)
	glUniform1f(locations[b"f_max"], 2*(f_max+1))
	glUniform1fv(locations[b"fold_map"], count+1, fold_map)
	glutTimerFunc(0, _unfold, ms()+T_FOLD)

def _unfold(t_end):
	folding = max(1. - 1.*(t_end - ms())/T_FOLD,
	              1./folding_widths[current])
	if folding > 1.:
		global interaction
		interaction = IDLE
		folding = 1.
		draw_values._reset()
	else:
		glutTimerFunc(20, _unfold, t_end)
	glUniform1f(locations[b"folding"], folding)
	glutPostRedisplay()


# dragging ##################################################################

di = 0
def drag(x1, y1):
	global x0, y0
	global di, parent

	i0, _ = ij(x0, 0.)
	i1, _ = ij(x1, 0.)
	di += i1-i0

	if di < 0:   # left move
		parent = right_child_parent(current)
		brother, _ = lr(parent)
	else: # right move
		parent = left_child_parent(current)
		_, brother = lr(parent)
	
	w = width(brother)
	if abs(di) > w:
		flip(parent)
		glUniform1fv(locations[b"fold_map"], count+1, fold_map)
		if di > 0: di -= w
		else:      di += w
	
	x0, y0 = x1, y1

def end_drag():
	global di
	l, r = lr(parent)
	wl, wr = width(l), width(r)
	if (di > 0. and di > wr/2.) or (di < 0. and -di > wl/2.):
		flip(parent)
		glUniform1fv(locations[b"fold_map"], count+1, fold_map)
	di = 0


# dragging two clusters ####################################################

di, dj = 0, 0
def drag_two(x1, y1):
	global x0, y0
	global di, dj
	global c_parent, d_parent
	
	i0, j0 = ij(x0, y0)
	i1, j1 = ij(x1, y1)
	di += i1-i0
	dj += j1-j0

	l, r = lr(current)
	c, d = selection_clusters
	assert c in childrens[l]
	assert d in childrens[r]
	
	# left side
	if di > 0: # right move
		c_parent = left_child_parent(c)
		_, brother = lr(c_parent)
	else:
		c_parent = right_child_parent(c)
		brother, _ = lr(c_parent)
	
	if c_parent not in childrens[l]:
		c_parent = None
	
	w = width(brother)
	if c_parent and abs(di) > w:
		flip(c_parent)
		glUniform1fv(locations[b"fold_map"], count+1, fold_map)
		if di > 0: di -= w
		else:      di += w

	# right side
	if dj > 0: # right move
		d_parent = left_child_parent(d)
		_, brother = lr(d_parent)
	else:
		d_parent = right_child_parent(d)
		brother, _ = lr(d_parent)
	
	if d_parent not in childrens[r]:
		d_parent = None
	
	w = width(brother)
	if d_parent and abs(dj) > w:
		flip(d_parent)
		glUniform1fv(locations[b"fold_map"], count+1, fold_map)
		if dj > 0: dj -= w
		else:      dj += w
	
	x0, y0 = x1, y1


def end_drag_two():
	di, dj = 0, 0


# hover #####################################################################

def hover(x, y):
	global highlight, current

	highlight[:] = leaves_index(x, y)
	for current, (i, a) in enumerate(spans):
		if all(i <= h <= a for h in highlight):
			break
	for folded in reversed(foldeds):
		if current in childrens[folded]:
			current = folded
			break


# keyboard interaction ######################################################
#############################################################################

freeze = False

def keyboard(c, x=0, y=0):
	global interaction, bending, cluster_threshold
	
	if c == CR:
		if interaction is IDLE:
			start_labeling()
			interaction = LABELING
	
	elif c == ESC:
		deselect()
	
	elif c in b"{}":
		if c == b"{":
			bending /= 1.05
		else:
			bending *= 1.05
		glUniform1f(locations[b"bending"], bending)
	
	elif c in b"[]<>":
		if c in b"[]":
			cluster_threshold += .01 if c == b"[" else -.01

		elif c in b"<>":
			if c == b"<":
				for _, _, s, _ in Z:
					if s > cluster_threshold:
						break
			else:
				for _, _, s, _ in reversed(Z):
					if s < cluster_threshold:
						break
			cluster_threshold = s
		
		cluster_threshold = min(max(0., cluster_threshold), 1.)
		draw_threshold._reset()
	
	elif c == 'f': # toggle full screen
		global WINDOW_SIZE
		if WINDOW_SIZE is None:
			WINDOW_SIZE = WIDTH, HEIGHT
			glutFullScreen()
		else:
			glutReshapeWindow(*WINDOW_SIZE)
			WINDOW_SIZE = None
	
	elif c == b's':
		print svg()
	
	elif c in b'mM':
		global freeze
		freeze = not freeze
		if freeze:
			glutMotionFunc(None)
			glutPassiveMotionFunc(None)
		else:
			glutMotionFunc(motion)
			glutPassiveMotionFunc(motion)

	
	elif c == b'q':
		sys.exit(0)

	glutPostRedisplay()


# labeling ##################################################################

def start_labeling():
	global caret, previous_label
	previous_label = labels.setdefault(current, u"")
	caret = len(previous_label)
	
	draw_labels._reset()
	glutKeyboardFunc(labeling_keyboard)
	glutSpecialFunc(labeling_special)
	glutSetCursor(GLUT_CURSOR_NONE)

def end_labeling():
	draw_labels._reset()
	glutKeyboardFunc(keyboard)
	glutSpecialFunc(None)
	glutSetCursor(GLUT_CURSOR_INHERIT)


BS, CR, ESC, DEL, NBSP, SHY = [unichr(k) for k in [8, 13, 27, 127, 160, 173]]

def labeling_keyboard(c, x, y):
	global interaction
	assert interaction == LABELING
	
	global labels, caret
	
	label = labels[current]
	
	if c in [CR, ESC]:
		if c == ESC:
			label = previous_label
		end_labeling()
		interaction = IDLE
	elif c == DEL:
		if caret > 0:
			label = label[:caret-1] + label[caret:]
			caret -= 1
	elif c == BS:
		label = label[:caret] + label[caret+1:]
	else:
		label = label[:caret] + c + label[caret:]
		caret += 1
	
	labels[current] = label
	
	glutPostRedisplay()

def labeling_special(i, x, y):
	global interaction
	assert interaction == LABELING
	
	global caret
	
	max_c = len(labels[current])
	if i == GLUT_KEY_RIGHT:
		caret = caret+1
	elif i == GLUT_KEY_LEFT:
		caret = caret-1
	elif i == GLUT_KEY_UP:
		caret = 0
	elif i == GLUT_KEY_DOWN:
		caret = max_c
	caret = max(0, min(caret, max_c))
	
	draw_labels._reset()
	
	glutPostRedisplay()


#############################################################################
# opengl initializations ####################################################
#############################################################################

def init_glut(argv):
	"""glut initialization."""
	glutInit(argv)
	glutInitWindowSize(WIDTH, HEIGHT)
	glutInitWindowPosition(100, 100)
#	glutInitDisplayString("rgba double samples=16 hidpi")
	glutInitDisplayString("rgba double samples=16")
	
	glutCreateWindow(argv[0].encode())

	glutReshapeFunc(reshape)
	glutDisplayFunc(display)
	glutKeyboardFunc(keyboard)
	glutMouseFunc(mouse)
	glutMotionFunc(motion)
	glutPassiveMotionFunc(motion)


uniforms = [b"bending", b"folding", b"f_min", b"f_max", b"fold_map", b"circle"]
locations = {}

def init_program():
	"""shader initialization."""
	vert_shader = glCreateShader(GL_VERTEX_SHADER)
	glShaderSource(vert_shader, b"""
		const int count = %s;
		uniform float bending;
		uniform float folding;
		uniform float f_min, f_max;
		uniform float fold_map[count+1];

		float _f(float u) {
			float ui = floor(u/2.);
			float uf = u/2. - ui;
			int i = int(ui);
			if(i<0)
				return 0.;
			if(i<count)
				return 2.*((1.-uf)*fold_map[i] + uf*fold_map[i+1]);
			return 2.*fold_map[count];
		}
		
		float fold(float u) {
			float f = _f(u);

			if(u < f_min)
				return f;

			float ff_min = _f(f_min);
			if(u < f_max)
				return ff_min + folding*(f-ff_min);

			float ff_max = _f(f_max);
			return ff_min + folding*(ff_max-ff_min) + (f-ff_max);
		}
		
		vec2 fold(vec2 p) {
			return vec2(fold(p.x), fold(p.y));
		}
		
		float _x(vec2 p) { return p.x+p.y; }
		float _y(vec2 p) { return p.x-p.y; }
		vec2 p(vec2 p)  {
			float x, y;
			if(_y(p) > 0.) {
				p = fold(p);
				x = _x(p);
				y = bending*log(1.+_y(p)/bending);
			} else {
				x = fold(_x(p)/2.)*2.;
				y = _y(p);
			}
			return vec2(x-fold(2.*float(count)), y);
		}
		
		void main() {
			gl_TexCoord[0] = gl_TextureMatrix[0] * gl_MultiTexCoord0;
			gl_Position = gl_ModelViewMatrix * gl_Vertex;
			gl_Position = gl_ProjectionMatrix * vec4(p(gl_Position.xy),
			                                          gl_Position.zw);
			gl_FrontColor = gl_Color;
		}
	""" % (count,))
	glCompileShader(vert_shader)
	if glGetShaderiv(vert_shader, GL_COMPILE_STATUS) != GL_TRUE:
		raise RuntimeError(glGetShaderInfoLog(vert_shader))
	
	frag_shader = glCreateShader(GL_FRAGMENT_SHADER)
	glShaderSource(frag_shader, b"""
		uniform bool circle;
		void main() {
			gl_FragColor = gl_Color;
			if(circle) {
				vec2 p = gl_TexCoord[0].st;
				float v = length(dFdx(p)+dFdy(p));
//				if(length(p) > 1.+v) discard;
				float alpha = clamp((1.-length(p))/v, 0., 1.);
				gl_FragColor.a *= alpha;
			}
		}
	""")
	glCompileShader(frag_shader)
	if glGetShaderiv(frag_shader, GL_COMPILE_STATUS) != GL_TRUE:
		raise RuntimeError(glGetShaderInfoLog(frag_shader))
	
	global program
	program = glCreateProgram()
	glAttachShader(program, vert_shader)
	glAttachShader(program, frag_shader)
	
	glLinkProgram(program)
	if glGetProgramiv(program, GL_LINK_STATUS) != GL_TRUE:
		raise RuntimeError(glGetProgramInfoLog(program))
	
	for uniform in uniforms:
		locations[uniform] = glGetUniformLocation(program, uniform)
	
	glUseProgram(program)
	glUniform1f(locations[b"bending"], bending)
	glUniform1f(locations[b"folding"], 1.)
	glUniform1fv(locations[b"fold_map"], count+1, fold_map)
	glUniform1i(locations[b"circle"], False)

	
def init_opengl():
	"""opengl state initialization."""
	glEnable(GL_BLEND)
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
	
	glClearColor(1., 1., 1., 1.)


def init_text():
	latin_1 = list(line.rstrip('\n') for line in open(os.path.join(os.path.dirname(sys.argv[0]), "latin-1.xo")))

	width, height = 6, 13
	n = len(latin_1)/height
	
	def b(s):
		"""bytes for a line."""
		for k in range((len(s)-1)/8+1):
			yield sum((1 << 7-i) if b != ' ' else 0
			          for i, b in enumerate(s[k*8:(k+1)*8]))
	def bs(lines):
		"""bytes for multiple lines."""
		return [o for line in lines for o in b(line)]
	
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
	
	global alphabet_v_id
	global alphabet_h_id
	alphabet_v_id = glGenLists(n)
	alphabet_h_id = glGenLists(n)
	
	for c in range(n):
		lines = latin_1[c*height:(c+1)*height]
		k = unichr(c+32)
		if k == NBSP: # nbsp
			lines = [" "*width] + ["o"*width]*(height-2) + [" "*width]
		elif k == SHY:
			lines = [" "*width] + [" "*(width-1) + "o"]*(height-2) + [" "*width]
		glNewList(alphabet_h_id+c, GL_COMPILE)
		glBitmap(width, height, 0, 0, width, 0, bs(reversed(lines)))
		glEndList()
		glNewList(alphabet_v_id+c, GL_COMPILE)
		glBitmap(height, width, 0, 0, 0, width, bs(zip(*lines)))
		glEndList()

	
def text_h(s, X=0., Y=0.):
	glRasterPos(0., 0.)
	glBitmap(0, 0, 0, 0, X, Y, None)
	for c in s.encode('latin-1', 'replace'):
		glCallList(alphabet_h_id+ord(c)-32)

def text_v(s, X=0., Y=0.):
	glRasterPos(0., 0.)
	glBitmap(0, 0, 0, 0, Y, X, None)
	for c in s.encode('latin-1', 'replace'):
		glCallList(alphabet_v_id+ord(c)-32)


#############################################################################
# svg output ################################################################
#############################################################################

def svg():
	from swit.graphics import scenegraph as sg
	from swit.graphics.scenegraph.xml import serializer

	def xy(i, j):
		return i+j, -bending*log(1.+(i-j)/bending)

	circle = sg.Circle(r=.85) #, stroke=sg.Color.gray)
	
	svg = sg.Group(stroke=sg.Color.black,
	               stroke_width=.2, stroke_opacity=.1,
	               children=[
		# clusters
		sg.Group(children=[
			sg.Polygon(points=list(xy(*p)
			                       for p in cluster_points(count+i)),
			           fill=sg.Color(s))
			for i, (_, _, s, _) in reversed(list(enumerate(Z)))
		]),

		# grid
#		sg.Group(fill=None, children=[
#			sg.Polyline(points=list(xy(*p) for p in line_points(2*i)))
#			for i in range(count+1)
#		]),

		# values
		sg.Group(children=[
			sg.Use(circle, stroke_opacity=1.,
			       transform=[sg.Translate(*xy(2*i+1, 2*j+1)),
			                  sg.Scale(1., bending*log((bending+2*(i-j)+1)/
			                                           (bending+2*(i-j)))),
			                  sg.Scale(sqrt(acos(X[leaves[i]][leaves[j]])))])
		for j in range(count)
		for i in range(j, count)
			if X[leaves[i]][leaves[j]] < similarity_threshold
		]),
		
		# authors
		sg.Group(text_anchor="end", children=[
			sg.Text(names[leaves[i]], font_size=4.,
			        transform=[sg.Translate(*xy(2*i+1, 2*i+1)),
			                   sg.Rotate(-90),
			                   sg.Translate(-2, 1.)])
			for i in range(count)
		]),
	])
	
	return serializer.serialize(svg).encode('utf-8')



#############################################################################
# main ######################################################################
#############################################################################

init_glut(sys.argv)
init_program()
init_opengl()
init_text()

sys.exit(glutMainLoop())
